{% load static i18n %}
<p>{% trans "Performs some checks inspired by the <a href='https://observatory.mozilla.org'>Mozilla's observatory</a>." %}</p>
<ul>
  {% for component in components %}
    <li class="">
      <h4><img class="djt-csp" src="{% static component.icon %}" alt="{{ component.level }}"> {{ component.title }}: {% if component.score > 0 %}+{% endif %}{{ component.score }}
      {% if component.reference_link %}<a href="{{ component.reference_link }}"><img alt="{% trans 'Help' %}" src="{% static "admin/img/icon-unknown.svg" %}"></a>{% endif %}
      </h4>
      {% if component.description %}<p>{{ component.description }}</p>{% endif %}
    {{ component.content }}
    </li>
  {% endfor %}
</ul>
